#!/usr/local/bin/php
<?php
/**
 * $Horde: whups/scripts/upgrades/2004-06-16_update_logs_table.php,v 1.3 2005/10/13 03:49:00 selsky Exp $
 *
 * This is a script to update the whups_logs table to the new format
 * effective 2004-06-16.
 */

@define('AUTH_HANDLER', true);
@define('HORDE_BASE', dirname(__FILE__) . '/../../..');
@define('WHUPS_BASE', HORDE_BASE . '/whups');

// Do CLI checks and environment setup first.
require_once HORDE_BASE . '/lib/core.php';
require_once 'Horde/CLI.php';

// Make sure no one runs this from the web.
if (!Horde_CLI::runningFromCLI()) {
    exit("Must be run from the command line\n");
}

// Load the CLI environment - make sure there's no time limit, init
// some variables, etc.
Horde_CLI::init();
$cli = &Horde_CLI::singleton();

require_once WHUPS_BASE . '/lib/base.php';
require_once WHUPS_BASE . '/lib/Driver.php';
$db = &$whups->_db;

// Fetch the current structure of the logs table.
$info = $db->tableInfo('whups_logs');

// These are the things we might need to do. We'll look at the table
// structure next to figure out which of them we still need to do.
$add_primary_key = true;
$add_transaction_id = true;
$migrate_comments = false;
$migrate_assignments = false;

// See what the structure says we should do.
foreach ($info as $field) {
    if ($field['name'] == 'log_id') {
        $add_primary_key = false;
    } elseif ($field['name'] == 'transaction_id') {
        $add_transaction_id = false;
    } elseif ($field['name'] == 'comment_id_log') {
        $migrate_comments = true;
    } elseif ($field['name'] == 'user_uid_action') {
        $migrate_assignments = true;
    }
}

// Tell the user what we're going to do.
if ($add_primary_key || $add_transaction_id || $migrate_comments || $migrate_assignments) {
    $cli->writeln('Will perform these actions:');
    if ($migrate_assignments) {
        $cli->writeln('- Convert assignment log entries.');
    }
    if ($add_primary_key) {
        $cli->writeln('- Add primary key.');
    }
    if ($add_transaction_id) {
        $cli->writeln('- Add transaction id.');
    }
    if ($migrate_comments) {
        $cli->writeln('- Convert comment log entries.');
    }
} else {
    $cli->writeln($cli->blue('Nothing to do!'));
    exit;
}

// Time to start converting.
$cli->writeln();

// Need to do this first so nothing else fails.
if ($migrate_comments) {
    fquery("ALTER TABLE whups_logs ALTER COLUMN comment_id_log DROP NOT NULL");
}

// Assignment log is easy, do that first.
if ($migrate_assignments) {
    $result = fquery("UPDATE whups_logs SET log_type = user_uid_action, log_value = user_uid_changed WHERE user_uid_action != ''");
    if (is_a($result, 'PEAR_Error')) {
        $cli->fatal($result->getMessage());
    }
}

// Add the primary key next, since having it will be helpful in the
// other steps.
if ($add_primary_key) {
    $entries = $db->getAll('SELECT ticket_id, log_timestamp, log_type, log_value FROM whups_logs', DB_FETCHMODE_ASSOC);
    if (is_a($entries, 'PEAR_Error')) {
        $cli->fatal($entries->getMessage());
    }

    // Now add the log_id field to the table.
    $result = fquery('ALTER TABLE whups_logs ADD COLUMN log_id INT');
    if (is_a($result, 'PEAR_Error')) {
        $cli->fatal($result->getMessage());
    }

    foreach ($entries as $entry) {
        $id = $db->nextId('whups_logs');
        $query = sprintf('UPDATE whups_logs SET log_id = %s WHERE ticket_id = %s AND log_timestamp = %s AND log_type %s AND log_value %s',
                         (int)$id,
                         (int)$entry['ticket_id'],
                         $db->quote($entry['log_timestamp']),
                         (is_null($entry['log_type']) ? 'IS ' : '= ') . $db->quote($entry['log_type']),
                         (is_null($entry['log_value']) ? 'IS ' : '= ') . $db->quote($entry['log_value']));

        $result = fquery($query);
        if (is_a($result, 'PEAR_Error')) {
            fquery('ALTER TABLE whups_logs DROP COLUMN log_id');
            $cli->fatal($result->getMessage());
        }
    }
}

// Add transaction ids before we throw away data that we'll need to do
// this.
if ($add_transaction_id) {
    // First step is to list all tickets so that we can then get each
    // one's history the way that Whups_Driver::getHistory() used to
    // figure it out. Then we'll assign each distinct update to each
    // ticket a transaction id.
    $tickets = $db->getAssoc('SELECT DISTINCT ticket_id, ticket_id FROM whups_logs');
    if (is_a($tickets, 'PEAR_Error')) {
        $cli->fatal($tickets->getMessage());
    }

    // Now add the transaction_id field to the table.
    $result = fquery('ALTER TABLE whups_logs ADD COLUMN transaction_id INT');
    if (is_a($result, 'PEAR_Error')) {
        $cli->fatal($result->getMessage());
    }

    foreach ($tickets as $id) {
        $history = get_history($id);
        if (is_a($history, 'PEAR_Error')) {
            $cli->fatal($history->getMessage());
        }

        foreach ($history as $log_ids) {
            if (!count($log_ids)) {
                continue;
            }

            $transaction_id = $db->nextId('whups_transactions');
            if (is_a($transaction_id, 'PEAR_Error')) {
                fquery('ALTER TABLE whups_logs DROP COLUMN transaction_id');
                $cli->fatal($transaction_id->getMessage());
            }

            $query = sprintf('UPDATE whups_logs SET transaction_id = %s WHERE log_id IN (%s)',
                             (int)$transaction_id,
                             implode(', ', $log_ids));
            $result = fquery($query);
            if (is_a($result, 'PEAR_Error')) {
                fquery('ALTER TABLE whups_logs DROP COLUMN transaction_id');
                $cli->fatal($result->getMessage());
            }
        }
    }
}

// Finish up by creating new entries for comments and then deleting
// the now irrelevant comment_id_log field.
if ($migrate_comments) {
    // Sanity check the data first.
    fquery('UPDATE whups_logs SET comment_id_log = NULL WHERE comment_id_log = 0');

    $entries = $db->getAll('SELECT DISTINCT ticket_id, transaction_id, comment_id_log, log_timestamp, user_id FROM whups_logs WHERE comment_id_log IS NOT NULL', DB_FETCHMODE_ASSOC);
    if (is_a($entries, 'PEAR_Error')) {
        $cli->fatal($entries->getMessage());
    }

    foreach ($entries as $entry) {
        $log_id = $db->nextId('whups_logs');
        if (is_a($log_id, 'PEAR_Error')) {
            $cli->fatal($log_id->getMessage());
        }

        $query = 'INSERT INTO whups_logs (log_id, transaction_id, ticket_id, log_timestamp, user_id, log_type, log_value) ' .
            sprintf('VALUES (%d, %d, %d, %d, %s, %s, %s)',
                    $log_id,
                    $entry['transaction_id'],
                    $entry['ticket_id'],
                    $entry['log_timestamp'],
                    $db->quote($entry['user_id']),
                    $db->quote('comment'),
                    $db->quote($entry['comment_id_log']));
        $result = fquery($query);
    }

    // We can now drop comment_id_log.
    fquery('ALTER TABLE whups_logs DROP COLUMN comment_id_log');

    // And we can finally get rid of all entries where log_type IS
    // NULL.
    fquery('DELETE FROM whups_logs WHERE log_type IS NULL');
}

$cli->writeln($cli->green('DONE!'));
exit;


/**
 * Get the history (old-style) for a ticket.
 */
function get_history($id)
{
    global $db;

    $query = 'SELECT log_id, user_id, log_timestamp, comment_id_log FROM whups_logs WHERE ticket_id = ' . (int)$id . ' ORDER BY log_timestamp';
    $rows = $db->getAll($query, DB_FETCHMODE_ASSOC);
    if (is_a($rows, 'PEAR_Error')) {
        return $rows;
    }

    $history = array();
    $transaction = null;
    $last_comment = 0;
    $last_time = 0;
    $last_user = null;
    foreach ($rows as $row) {
        if ($row['user_id'] != $last_user) {
            // If the userId changed, we know we have a new
            // transaction.
            $transaction = $row['log_timestamp'];
        } elseif ($last_comment !== $row['comment_id_log']) {
            $transaction = $row['log_timestamp'];
        } else {
            // Start a new transaction if there's more than a
            // single-second difference between this entry and the
            // previous one.
            if (abs($last_time - $row['log_timestamp']) > 1) {
                $transaction = $row['log_timestamp'];
            }
        }

        // Track the last row's userId, commentId, and timestamp.
        $last_time = $row['log_timestamp'];
        $last_user = $row['user_id'];
        $last_comment = $row['comment_id_log'];

        if (empty($history[$transaction])) {
            $history[$transaction] = array();
        }
        $history[$transaction][] = $row['log_id'];
    }

    return $history;
}

function fquery($sql)
{
    global $db, $cli;
    $result = $db->query($sql);
    if (is_a($result, 'PEAR_Error')) {
        $cli->fatal($result->getMessage() . ": " . $result->getUserInfo());
    }

    return $result;
}
